program heateqn
    implicit none

    integer, parameter :: filenamelen=256
    integer, parameter :: maxiters=500000

    ! defaults 
    integer :: numcvs=500
    integer :: method=1
    character(len=20), dimension(3) :: methodnames = (/ "Gauss-Seidel    ",  &
                                    "Red-Black       ", "Jacobi Iteration" /)
    double precision :: tol=1.d-6
    character(len=filenamelen) :: outfile, logfile
    character(len=filenamelen) :: cmdstring, message

    double precision, dimension(maxiters) :: err, time

    ! parameters of the problem
    double precision, parameter :: pi = 4.*atan(1.d0)
    double precision, parameter :: Tleft = 1.d0
    double precision, parameter :: Tright = 1.d0
    double precision, parameter :: Length = 2.d0*pi
    double precision, parameter :: Lambda = 1.d0
    double precision, allocatable, dimension(:) :: leftband, diagonal, &
                                                   rightband, x, soln, rhs
    integer :: i
    double precision :: dx, L1, L2, L_inf, T_exact, errsum
    integer :: niters=1
    double precision :: resid
    integer :: timer
    double precision:: computetime

    !! get command line arguments using namelists
    outfile="out-serial.txt"
    logfile="log-serial.txt"
    call  getargs(method, numcvs, tol, outfile, logfile)

    call tick(timer)
    !Variable initialization
    allocate(leftband(numcvs+2))
    allocate(rightband(numcvs+2))
    allocate(diagonal(numcvs+2))
    allocate(soln(numcvs+2))
    allocate(x(numcvs+2))
    allocate(rhs(numcvs+2))

    !calculate the length of each cv
    dx = Length/real(numcvs)

    !define the distance to the center of each c.v.
    x(1) = -dx/2.d0
    do i=2,numcvs+2
        x(i) = x(i-1) + dx
    end do

    forall (i=2:numcvs+1)
        leftband(i) = -lambda/dx
        diagonal(i) = 2.*lambda/dx
        rightband(i)= -lambda/dx
        rhs(i) = cos(x(i))*dx
    end forall

    !solve the tridiagonalonal linear system using the Thomas algorithm
    select case (method)
        case (1)
            call gaussSeidel(leftband,diagonal,rightband,rhs,soln,&
                             Tleft,Tright,tol,maxiters,niters,err,time)
        case (2)
            call redBlackGaussSeidel(leftband,diagonal,rightband,rhs,soln,&
                             Tleft,Tright,tol,maxiters,niters,err,time)
        case (3)
            call jacobi(leftband,diagonal,rightband,rhs,soln,&
                             Tleft,Tright,tol,maxiters,niters,err,time)
    end select

    if (niters >= maxiters .and. err(niters) > tol) then
        print *,'Warning: did not converge to specified tolerance ', tol
        print *,'in specified number of iterations ', maxiters
    endif

    !Calculate the three errors, L1, L2, and L_inf
    L1 = 0.d0
    L2 = 0.d0
    L_inf = 0.d0
    do i=2,numcvs+1
        T_exact = cos(x(i))
        resid = dabs(soln(i)-T_exact)
        L1 = L1 + resid
        L2 = L2 + resid**2
        if (resid > L_inf) L_inf = resid
    end do
    computetime = tock(timer)

    print *,'Computed solution'
    print *,'  Num Control volumes = ', numcvs, '.'
    print *,'  Using method ', methodnames(method), '.'
    print *,'  Time taken = ', computetime, '.'
    print *,'  Number of iterations = ', niters
    print *,'  Error: L1 = ', L1, ' L2 = ', L2, ' L_inf = ', L_inf
    print *,'  Outputing solution to ', trim(outfile)

    ! write the computed temperatures to a file
    open(file=trim(outfile),unit=21)
    write(21,*) '#'
    write(21,*) '# numcvs = ',numcvs
    write(21,*) '#      x(i)          T(i)'
    do i=2,numcvs+1
        write(21,*) x(i), soln(i)
    end do
    close(unit=21)

    print *,'  Outputing convergence history to ', trim(logfile)
    open(file=trim(logfile),unit=21)
    write(21,*) '#'
    write(21,*) '# Convergence History'
    write(21,*) '#      i        rms residual     time'
    do i=1,niters
        write(21,*) i, err(i), time(i)
    end do
    close(unit=21)

contains
subroutine boundaryConditions(soln, Tleft, Tright)
    implicit none
    double precision, intent(inout), dimension(:) :: soln
    double precision, intent(in)  :: Tleft, Tright
    integer :: n

    n = size(soln,1) - 2

    soln(1)   = Tleft
    soln(n+2) = Tright
end subroutine boundaryConditions

function rmsResidual(x, l, d, r, rhs) result(resid)
    implicit none
    double precision :: resid
    double precision, dimension(:), intent(in) :: x, l, d, r, rhs
    integer :: n

    n = size(soln,1)-2
    resid = sum((rhs(2:n+1) - l(2:n+1)*x(1:n)-d(2:n+1)*x(2:n+1) - r(2:n+1)*x(3:n+2))**2)

    resid = sqrt(resid)/n
end function 

! This subroutine solves a tri-diagonalonal linear system using 
! Gauss-Seidel
subroutine gaussSeidel(l,d,r,rhs,x,Tleft,Tright,tol,maxiters,niters,errors,times)
    implicit none
    double precision, dimension(:), intent(in) :: l, d, r, rhs
    double precision, dimension(:), intent(inout) :: x
    double precision, intent(in) :: Tleft, Tright
    double precision, intent(in) :: tol
    integer, intent(in) :: maxiters
    double precision, intent(out), dimension(maxiters) :: errors
    double precision, intent(out), dimension(maxiters) :: times
    integer, intent(out) :: niters
    integer :: n
    integer :: iter
    integer :: timer
    
    n = size(x,1)-2
    iter = 1
    call tick(timer)

    ! Generate the initial guess
    x(2:n+1) = rhs(2:n+1)/d(2:n+1)
    
    ! impose boundary conditions
    call boundaryConditions(x, Tleft, Tright)

    ! perform the GS iterations
    do 
        ! update
        do i=2,n+1
            x(i)=(rhs(i)-l(i)*x(i-1)-r(i)*x(i+1))/d(i)
        enddo
        
        ! impose BCs
        call boundaryConditions(x, Tleft, Tright)    
    
        ! find residual
        errsum = rmsResidual(x, l, d, r, rhs)
        errors(iter) = errsum
        times(iter)  = tock(timer)
        
        iter = iter + 1
        ! exit loop if converged, or too many iters
        if (errsum < tol) exit
        if (iter > maxiters) exit
    end do

    niters = iter-1

end subroutine gaussSeidel

! This subroutine solves a tri-diagonalonal linear system using 
! reordered (Red-Black) Gauss-Seidel
subroutine redBlackGaussSeidel(l,d,r,rhs,x,Tleft,Tright,tol,maxiters,niters,errors,times)
    implicit none
    double precision, dimension(:), intent(in) :: l, d, r, rhs
    double precision, dimension(:), intent(inout) :: x
    double precision, intent(in) :: Tleft, Tright
    double precision, intent(in) :: tol
    integer, intent(in) :: maxiters
    double precision, intent(out), dimension(maxiters) :: errors
    double precision, intent(out), dimension(maxiters) :: times
    integer, intent(out) :: niters
    integer :: n
    integer :: iter
    integer :: timer
    
    n = size(x,1)-2
    iter = 1
    call tick(timer)

    ! Generate the initial guess
    x(2:n+1) = rhs(2:n+1)/d(2:n+1)
    
    ! impose boundary conditions
    call boundaryConditions(x, Tleft, Tright)

    ! perform the GS iterations
    do 
        ! update - Red
        forall (i=2:n+1:2)
            x(i)=(rhs(i)-l(i)*x(i-1)-r(i)*x(i+1))/d(i)
        end forall
        
        ! impose BCs
        call boundaryConditions(x, Tleft, Tright)    

        ! update - Black
        forall (i=3:n+1:2)
            x(i)=(rhs(i)-l(i)*x(i-1)-r(i)*x(i+1))/d(i)
        end forall
        
        ! impose BCs
        call boundaryConditions(x, Tleft, Tright)    
    
        ! find residual
        errsum = rmsResidual(x, l, d, r, rhs)
        errors(iter) = errsum
        times(iter)  = tock(timer)
        
        iter = iter + 1
        ! exit loop if converged, or too many iters
        if (errsum < tol) exit
        if (iter > maxiters) exit
    end do

    niters = iter-1
end subroutine redBlackGaussSeidel

subroutine swappointers(x,y)
    implicit none
    double precision, dimension(:), pointer, intent(inout) :: x,y
    double precision, dimension(:), pointer :: p1, p2

    p1 => x
    p2 => y
    x => p2
    y => p1
end subroutine swappointers

! This subroutine solves a tri-diagonalonal linear system using 
! Jacobi iteration
subroutine jacobi(l,d,r,rhs,x,Tleft,Tright,tol,maxiters,niters,errors,times)
    implicit none
    double precision, dimension(:), intent(in) :: l, d, r, rhs
    double precision, dimension(:), intent(inout), target :: x
    double precision, intent(in) :: Tleft, Tright
    double precision, intent(in) :: tol
    integer, intent(in) :: maxiters
    double precision, intent(out), dimension(maxiters) :: errors
    double precision, intent(out), dimension(maxiters) :: times
    integer, intent(out) :: niters
    integer :: n
    integer :: iter
    integer :: timer
    double precision, dimension(:), allocatable, target :: work
    double precision, dimension(:), pointer :: xold, xnew
    
    n = size(x,1)-2
    iter = 1
    call tick(timer)

    ! allocate temporary array
    
    allocate(work(n+2))
    xold => x
    xnew => work

    ! Generate the initial guess
    xold(2:n+1) = rhs(2:n+1)/d(2:n+1)
    
    ! impose boundary conditions
    call boundaryConditions(xold, Tleft, Tright)

    ! perform the Jacobi iterations
    do 
        ! update 
        forall (i=2:n+1)
            xnew(i)=(rhs(i)-l(i)*xold(i-1)-r(i)*xold(i+1))/d(i)
        end forall
        
        ! impose BCs
        call boundaryConditions(xnew, Tleft, Tright)    

        ! find residual
        errsum = rmsResidual(xnew, l, d, r, rhs)
        errors(iter) = errsum
        times(iter)  = tock(timer)
        
        iter = iter + 1
        call swappointers(xold, xnew)

        ! exit loop if converged, or too many iters
        if (errsum < tol) exit
        if (iter > maxiters) exit
    end do
    niters = iter-1

    ! if we've iterated an odd number of times, then 
    ! the real data is in work; we need to copy that
    ! back to x
    if (mod(iter,2) == 1) then
        x = work
    endif

    nullify(xold,xnew)
    deallocate(work)
end subroutine jacobi

subroutine getargs(method, numcvs, tol, outfile, logfile)
    implicit none
    integer, intent(inout) :: method
    integer, intent(inout) :: numcvs
    double precision, intent(inout) :: tol
    character(len=256), intent(inout) :: outfile
    character(len=256), intent(inout) :: logfile
    namelist /args/ method,numcvs,tol,outfile,logfile
    integer :: arg, nargs
    integer :: iostatus

    nargs = command_argument_count()

    do arg=1,nargs
        call get_command_argument(nargs, cmdstring)
        if (len_trim(cmdstring) == 0) exit

        cmdstring="&args "//trim(cmdstring)//" /"
        read(cmdstring,NML=args,iostat=iostatus,iomsg=message)
        if (iostatus /= 0) then
            print *, 'Error parsing command line arguments', cmdstring
            print *, message
            stop 
        endif
    end do

    if (method < 1 .or. method > 3) then
        print *, 'Error: cannot use method ', method
        stop
    endif

    if (numcvs < 10 .or. numcvs > 100000) then
        print *, 'Error: cannot use numcvs', numcvs
        stop
    endif

    if (tol < 1.d-12 .or. tol > 1.) then
        print *, 'Error: cannot use tol', tol
        stop
    endif

    open(unit=22,file=outfile,iostat=iostatus)
    if (iostatus /= 0) then
        print *, 'Error: cannot use outfile ', outfile
        stop
    endif
    close(22)

    open(unit=22,file=logfile,iostat=iostatus)
    if (iostatus /= 0) then
        print *, 'Error: cannot use logfile ', logfile
        stop
    endif
    close(22)

end subroutine getargs

subroutine tick(t)
    implicit none
    integer, intent(OUT) :: t

    call system_clock(t)
end subroutine tick

    ! returns time in seconds from now to time described by t
double precision function tock(t)
    implicit none
    integer, intent(in) :: t
    integer :: now, clock_rate

    call system_clock(now,clock_rate)

    tock = real(now - t)/real(clock_rate)
end function tock

end program heateqn
