#include <stdio.h>
#include <math.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <fenv.h>

void tick(struct timeval *t);
double tock(struct timeval *t);

/*
 * Based on a program written by Dr. Seth Dworkin on January 30, 2010.
 * This program solves problem of 1D heat conduction in a straight rod with 
 * Dirichlet boundary conditions and a cos(x) source term using finite volumes
 * and solves the linear system using Gauss-Seidel iteration, Red-Black Gauss-
 * Seidel iteration, or Jacobi iteration.
 */

#define MAXFILENAME 128
int get_options(int argc, char **argv, int *numcvs, double *tol, int *method, 
                char *outfilename, char *convfilename);

#define GAUSSSEIDEL   1
#define RBGAUSSSEIDEL 2
#define JACOBI        3

#define SQUARE(x) (x)*(x)
/*
 * Routines for doing the linear solver of a tridiagonal system; 
 *  A general residual calculator,
 *  A Gauss-Seidel iteration routine,
 *  A reordered (Red-Black) Gauss-Seidel iteration routine,
 *  A Jacobi Iteration routine.
 */
double rmsResidual(int n, double *x, double *l, double *d, double *r, 
                   double *rhs) {
    double err;

    err = 0.;
    for (int i=1; i<=n; i++) {
        err += SQUARE(rhs[i]-l[i]*x[i-1]-d[i]*x[i]-r[i]*x[i+1]);
    }
    err = sqrt(err)/n;

    return err;
}
/* solve a tridiagonal linear system using Gauss-Seidel */

void boundaryConditions(int n, double *soln, double Tleft, double Tright) {
    soln[0]   = Tleft;
    soln[n+1] = Tright;
    return;
}

void gaussSeidel(int n, double tol, double *l, double *d, double *r, 
                 double *rhs, double *x, const int maxiters, int *iters, 
                 double *errors, double *times, double Tleft, double Tright) {

    double err;
    struct timeval start; 

    tick(&start);
    /* initial guess: */
    for (int i=1; i<=n; i++)
        x[i] = rhs[i]/d[i];

    *iters = 0;
    err = 1.e+9;
    do {
        /* update x */
        for (int i=1; i<=n; i++) {
            x[i] = (rhs[i] - l[i]*x[i-1] - r[i]*x[i+1])/d[i];
        }

        boundaryConditions(n,x,Tleft,Tright);

        /* update RMS error */
        err = rmsResidual(n,x,l,d,r,rhs);

        errors[*iters] = err;
        times[*iters] = tock(&start);
        (*iters)++;
    } while (err > tol && *iters < maxiters);

    return;
}    

/* solve a tridiagonal linear system using Red-Black Gauss-Seidel */

void gaussSeidelRedBlack(int n, double tol, double *l, double *d, double *r, 
                         double *rhs, double *x, const int maxiters, 
                         int *iters, double *errors, double *times, 
                         double Tleft, double Tright) {
    double err;
    struct timeval start;
       
    tick(&start);
    /* initial guess: */
    for (int i=1; i<=n; i++)
        x[i] = rhs[i]/d[i];

    *iters = 0;
    err = 1.e+9;
    do {
        /* Red */
        for (int i=1; i<=n; i+=2) {
            x[i] = (rhs[i] - l[i]*x[i-1] - r[i]*x[i+1])/d[i];
        }
        boundaryConditions(n,x,Tleft,Tright);

        /* Black */
        for (int i=2; i<=n; i+=2) {
            x[i] = (rhs[i] - l[i]*x[i-1] - r[i]*x[i+1])/d[i];
        }
        boundaryConditions(n,x,Tleft,Tright);

        /* update RMS error */
        err = rmsResidual(n,x,l,d,r,rhs);

        errors[*iters] = err;
        times[*iters] = tock(&start);
        (*iters)++;
    } while (err > tol && *iters < maxiters);

    return;
}    

/* solve a tridiagonal linear system using Jacobi */

void jacobi(int n, double tol, double *l, double *d, double *r, 
            double *rhs, double *x, const int maxiters, int *iters, 
            double *errors, double *times, double Tleft, double Tright) {

    double err;
    double *work, *xold, *xnew;
    struct timeval start;

    work = (double *)malloc((n+2)*sizeof(double));
    xnew = work;
    xold = x;
       
    /* initial guess: */
    for (int i=1; i<=n; i++)
        xold[i] = rhs[i]/d[i];

    *iters = 0;
    err = 1.e+9;
    tick(&start);
    do {
        /* update x */
        for (int i=1; i<=n; i++) {
            xnew[i] = (rhs[i] - l[i]*xold[i-1] - r[i]*xold[i+1])/d[i];
        }

        boundaryConditions(n,xnew,Tleft,Tright);

        /* update RMS error */
        err = rmsResidual(n,xnew,l,d,r,rhs);

        errors[*iters] = err;
        times[*iters] = tock(&start);
        (*iters)++;

        double *p = xold;
        xold = xnew;
        xnew = p; 
    } while (err > tol && *iters < maxiters);

    /* if we've done an odd number of iterations, the most recent data -
     * which is now pointed to by xold - must be copied back to x.
     */

    if ((*iters % 2) == 1) {
        memcpy(x, xold, (n+2)*sizeof(double));
    }        
    free(work);
    return;
}    

/*
 * The main loop.
 */
int main(int argc, char **argv) {
    int numcvs;
    double tol;
    int method;
    char outfilename[MAXFILENAME]; 
    char convfilename[MAXFILENAME]; 

    const double lambda = 1.e0;
    const double pi=4.*atan(1.0);
    const double length=2.*pi;
    const double Tleft =1.0;
    const double Tright=1.0;

    double dx;
    double *x, *leftband, *rightband, *diagonal, *rhs;
    double *soln;
    double *errors;
    double *times;

    struct timeval compute;
    double computetime;

    const int MAXITERS=500000;
    int iters;
               
    double l1=0., l2=0., linf=0.;

    char *methodnames[4];

    methodnames[GAUSSSEIDEL]  = "Gauss-Seidel Iteration";
    methodnames[RBGAUSSSEIDEL]= "Red-Black Gauss-Seidel Iteration";
    methodnames[JACOBI]       = "Jacobi Iteration";

    if (get_options(argc, argv, &numcvs, &tol, &method, outfilename, 
                    convfilename)) 
        return 0;
    
    x = (double *)malloc((numcvs+2)*(sizeof(double)));
    leftband = (double *)malloc((numcvs+2)*(sizeof(double)));
    rightband= (double *)malloc((numcvs+2)*(sizeof(double)));
    diagonal = (double *)malloc((numcvs+2)*(sizeof(double)));
    rhs      = (double *)malloc((numcvs+2)*(sizeof(double)));
    soln     = (double *)malloc((numcvs+2)*(sizeof(double)));
    errors   = (double *)malloc(MAXITERS*(sizeof(double)));
    times    = (double *)malloc(MAXITERS*(sizeof(double)));
    if (!x||!leftband||!rightband||!diagonal||!rhs||!soln||!errors||!times) {
        fprintf(stderr,"%s: Error allocating memory.\n",argv[0]);
        fprintf(stderr,"\tExiting.\n");
        exit(-1);
    }

    /* positions are the control volume centers */
    dx = length / numcvs;
    x[0] = -dx/2;
    for (int i=1; i<numcvs+2; i++) 
        x[i] = x[i-1]+dx;         

    /* interior points */
    for (int i=1; i<=numcvs; i++) {
        leftband [i] = -lambda/dx;
        diagonal [i] = 2.*lambda/dx;
        rightband[i] = -lambda/dx;
        rhs      [i] = cos(x[i])*dx;
    }

    /* impose BCs: */
    boundaryConditions(numcvs, soln, Tleft, Tright);

    tick(&compute);
    switch (method) {
        case JACOBI:
            jacobi(numcvs, tol, 
                       leftband, diagonal, rightband, rhs, soln, 
                       MAXITERS, &iters, errors, times, Tleft, Tright);
            break;
        case RBGAUSSSEIDEL:
            gaussSeidelRedBlack(numcvs, tol, 
                       leftband, diagonal, rightband, rhs, soln, 
                       MAXITERS, &iters, errors, times, Tleft, Tright);
            break;
        default:
        case GAUSSSEIDEL:
            gaussSeidel(numcvs, tol, 
                       leftband, diagonal, rightband, rhs, soln, 
                       MAXITERS, &iters, errors, times, Tleft, Tright);
            break;
    }
    computetime = tock(&compute);

    if (iters >= MAXITERS && errors[iters-1] > tol) {
        fprintf(stderr,"%s: Warning: Did not converge to specified tolerance (%lf)\n", argv[0], tol);
        fprintf(stderr,"\tin specifed number of iterations(%d).\n", MAXITERS);
    }

    for (int i=1; i<=numcvs; i++) {
        double exactsoln = cos(x[i]);
        double resid = fabs(soln[i] - exactsoln);

        l1 += resid;
        l2 += resid*resid;
        linf = ( resid > linf ? resid : linf );
    }

    printf("Computed solution\n");
    printf("  Num Control volumes = %d.\n", numcvs);
    printf("  Time taken = %lf.\n", computetime);
    printf("  Number of iterations = %d.\n", iters);
    printf("  Error: L1 = %lf, L2 = %lf, L_infty = %lf.\n", l1, l2, linf);
    printf("  Outputting to file %s.\n", outfilename);
    printf("  Using %s.\n", methodnames[method]);

    FILE *file;
    file = fopen(outfilename,"w");
    fprintf(file,"# Heat conduction test\n# X\t\tT\n");
    for (int i=1; i<=numcvs; i++) {
        fprintf(file,"%lf\t%lf\n",x[i], soln[i]);
    }
    fclose(file);

    printf("  Outputting convergence history to file %s.\n", convfilename);
    file = fopen(convfilename,"w");
    fprintf(file,"# Heat conduction test\n# Iter\t\tError\t\tTime\n");
    for (int i=0; i<iters; i++) {
        fprintf(file,"%d\t%lf\t%lf\n",i, errors[i], times[i]);
    }
    fclose(file);

    free(x);
    free(leftband);
    free(rightband);
    free(diagonal);
    free(rhs);
    free(soln);
    free(errors);
    free(times);

    return 0;
}


/*
 *  Utility routines for doing timing, getting command-line 
 *  options.
 */
void tick(struct timeval *t) {
    gettimeofday(t, NULL);
}

/* returns time in seconds from now to time described by t */
double tock(struct timeval *t) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (double)(now.tv_sec - t->tv_sec) +
             ((double)(now.tv_usec - t->tv_usec)/1000000.);
}


int get_options(int argc, char **argv, int *numcvs, double *tol, int *method, 
                char *outfilename, char *convfilename) {
                                        
    struct option long_options[] = {
        {"ncvs",     required_argument, 0, 'n'},
        {"tol",      required_argument, 0, 't'},
        {"out",      required_argument, 0, 'o'},
        {"log",      required_argument, 0, 'l'},
        {"method",   required_argument, 0, 'm'},
        {"help",     no_argument, 0, 'h'},
        {0, 0, 0, 0}};

    char c;
    int option_index;
    int tempint;
    double tempflt;
    int defaultnpts = 500;
    double defaulttol = 1.e-6;
    FILE *tst;
    char *defaultofname="T_of_x.dat";
    char *defaultlfname="linsysres3a.dat";
    int defaultmethod=GAUSSSEIDEL;

    *numcvs = defaultnpts;
    *tol = defaulttol;
    *method = defaultmethod;
    strncpy(outfilename, defaultofname, MAXFILENAME-1);
    strncpy(convfilename, defaultlfname, MAXFILENAME-1);

    while (1) { 
        c = getopt_long(argc, argv, "n:t:o:l:h", long_options, 
                        &option_index);
        if (c == -1) break;

        switch (c) { 
            case 0: if (long_options[option_index].flag != 0)
                    break;

            case 'n': tempint = atoi(optarg);
                      if (tempint < 1 || tempint > 500000) {
                          fprintf(stderr,
                                  "%s: Cannot use # of control volumes%s;\n",
                                  argv[0], optarg);
                          fprintf(stderr,"  Using %d\n", defaultnpts);
                          *numcvs = defaultnpts;
                      } else {
                          *numcvs = tempint;
                      }
                      break;

            case 'm': tempint = atoi(optarg);
                      if (tempint < 1 || tempint > 3) {
                          fprintf(stderr,
                                  "%s: Cannot use method number %s;\n",
                                  argv[0], optarg);
                          fprintf(stderr,"  Using %d\n", defaultmethod);
                          *method = defaultmethod;
                      } else {
                          *method = tempint;
                      }
                      break;

            case 't': tempflt = atof(optarg);
                      if (tempflt < 0.) {
                          fprintf(stderr,
                                  "%s: Cannot use convergence tolerance %s;\n",
                                  argv[0], optarg);
                          fprintf(stderr,"  Using %lf\n", defaulttol);
                          *tol= defaulttol;
                      } else {
                          *tol = tempflt;
                      }
                      break;

            case 'o': strncpy(outfilename, optarg, MAXFILENAME-1);
                      if (!(tst = fopen(outfilename,"w"))) {
                          fprintf(stderr,
                                  "Cannot use output filename %s;\n",
                                  outfilename);
                          fprintf(stderr, "  Using %s\n",defaultofname);
                          strncpy(outfilename, defaultofname, MAXFILENAME-1);
                      } else
                          fclose(tst);
                      break; 

            case 'l': strncpy(convfilename, optarg, MAXFILENAME-1);
                      if (!(tst = fopen(convfilename,"w"))) {
                          fprintf(stderr,
                                  "Cannot use convergence log filename %s;\n",
                                  convfilename);
                          fprintf(stderr, "  Using %s\n",defaultlfname);
                          strncpy(convfilename, defaultlfname, MAXFILENAME-1);
                      } else
                          fclose(tst);
                      break; 

            case 'h':
                  puts("Options: ");
                  puts("    --ncvs=N   (-n N): Set the number of grid cells in x direction.");
                  puts("    --tol=X    (-t X): Set the convergence tolerance.");
                  puts("    --out=S    (-o S): Set the output filename.");
                  puts("    --log=S    (-l S): Set the convergence log filename.");
                  puts("    --method=N (-m N): Set the method for solving the system.");
                  printf("                        %d == Gauss-Siedel,\n", GAUSSSEIDEL);
                  printf("                        %d == Red-Black Gauss-Siedel,\n", RBGAUSSSEIDEL);
                  printf("                        %d == Jacobi iteration.\n", JACOBI);
                  puts("");
                  return +1;

            default: printf("Invalid option %s\n", optarg);
                 break;
        }
    }
    return 0;
}
